import platform
from collections import Counter

import torch
import numpy as np
import torchvision
import matplotlib
import matplotlib.pyplot as plt
from torch.hub import download_url_to_file
import os
import pickle
import torch.utils.data
import torch.nn.functional as F
from torch.utils.data import Subset
from tqdm import tqdm
import sklearn.model_selection
import torch.utils.data


plt.rcParams["figure.figsize"] = (15, 5)
plt.style.use('dark_background')

LEARNING_RATE = 1e-4
BATCH_SIZE = 128
MAX_LEN = 200
TRAIN_TEST_SPLIT = 0.7
DEVICE = 'cpu'

if torch.cuda.is_available():
    DEVICE = 'cuda'
    MAX_LEN = 0

class Dataset(torch.utils.data.Dataset):
    def __init__(self):
        super().__init__()
        path_dataset = '../data/Fruits28.pkl'
        if not os.path.exists(path_dataset):
            pass
            os.makedirs('../data', exist_ok=True)
            download_url_to_file(
                'http://share.yellowrobot.xyz/1645110979-deep-learning-intro-2022-q1/Fruits28.pkl',
                path_dataset,
                progress=True
            )
        with open(path_dataset, 'rb') as fp:
            X, Y, self.labels = pickle.load(fp)
        self.Y_idx = Y

        Y_counter = Counter(Y)
        Y_counts = np.array(list(Y_counter.values()))
        self.Y_weights = (1.0 / Y_counts) * np.sum(Y_counts)

        X = torch.from_numpy(np.array(X).astype(np.float32))
        self.X = X.permute(0, 3, 1, 2)
        self.input_size = self.X.size(-1)
        Y = torch.LongTensor(Y)
        self.Y = F.one_hot(Y)

    def __len__(self):
        return len(self.X)

    def __getitem__(self, idx):
        x = self.X[idx]
        y = self.Y[idx]

        return x, y


dataset_full = Dataset()
train_test_split = int(len(dataset_full) * TRAIN_TEST_SPLIT)

idxes_train, idxes_test = sklearn.model_selection.train_test_split(
    np.arange(len(dataset_full)),
    train_size=train_test_split,
    test_size=len(dataset_full) - train_test_split,
    stratify=dataset_full.Y_idx,
    random_state=0
)

# For debugging
if MAX_LEN:
    idxes_train = idxes_train[:MAX_LEN]
    idxes_test = idxes_test[:MAX_LEN]

dataset_train = Subset(dataset_full, idxes_train)
dataset_test = Subset(dataset_full, idxes_test)

dataloader_train = torch.utils.data.DataLoader(
    dataset=dataset_train,
    batch_size=BATCH_SIZE,
    shuffle=True,
    drop_last=(len(dataset_train) % BATCH_SIZE == 1)
)

dataloader_test = torch.utils.data.DataLoader(
    dataset=dataset_test,
    batch_size=BATCH_SIZE,
    shuffle=False,
    drop_last=(len(dataset_test) % BATCH_SIZE == 1)
)

def get_positional_embeddings(sequence_length, d):
    result = torch.ones(sequence_length, d).to(DEVICE)
    for i in range(sequence_length):
        for j in range(d):
            result[i][j] = np.sin(i / (10000 ** (j / d))) if j % 2 == 0 else np.cos(i / (10000 ** ((j - 1) / d)))
    return result

class PositionalEncoding(torch.nn.Module):
    def __init__(self, num_embeddings, embedding_dim):
        super().__init__()

        pe = torch.zeros(num_embeddings, embedding_dim).to(DEVICE)
        position = torch.arange(0, num_embeddings, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, embedding_dim, 2).float() * (-np.log(10000.0) / embedding_dim))
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        self.pe = pe

    def forward(self, idxes):
        return self.pe[idxes, :]

class TransformerLayer(torch.nn.Module):
    def __init__(self, hidden_size, n_heads=2):
        super().__init__()

        self.hidden_size = hidden_size
        self.n_heads = n_heads
        self.project_k = torch.nn.Linear(in_features=hidden_size, out_features=hidden_size)
        self.project_q = torch.nn.Linear(in_features=hidden_size, out_features=hidden_size)
        self.project_v = torch.nn.Linear(in_features=hidden_size, out_features=hidden_size)

        self.ff = torch.nn.Sequential(
            torch.nn.Linear(in_features=hidden_size, out_features=hidden_size),
            torch.nn.GELU(),
            torch.nn.Linear(in_features=hidden_size, out_features=hidden_size)
        )

        self.norm_1 = torch.nn.LayerNorm(normalized_shape=hidden_size)
        self.norm_2 = torch.nn.LayerNorm(normalized_shape=hidden_size)

    def forward(self, x):
        batch_size = x.size(0) # x.shape (B, seq, HIDDEN_SIZE)
        seq_size = x.size(1)

        # shape (B, seq, HIDDEN_SIZE)
        k = self.project_k.forward(x)
        q = self.project_q.forward(x)
        v = self.project_v.forward(x)

        # shape (B, seq, Heads, HIDDEN_SIZE/Heads)
        # shape (B, Heads, seq, HIDDEN_SIZE/Heads)
        k = k.view(batch_size, seq_size, self.n_heads, int(self.hidden_size/self.n_heads)).transpose(1, 2)
        q = q.view(batch_size, seq_size, self.n_heads, int(self.hidden_size/self.n_heads)).transpose(1, 2)
        v = v.view(batch_size, seq_size, self.n_heads, int(self.hidden_size/self.n_heads)).transpose(1, 2)

        atten_raw = q @ k.transpose(-1, -2) / np.sqrt(x.size(-1))

        atten = torch.softmax(atten_raw, dim=-1)
        atten = atten.masked_fill(((atten > 0) == False), value=0.0)
        out = atten @ v

        out = out.transpose(1, 2)
        out = out.contiguous().view(batch_size, seq_size, self.hidden_size)
        atten = atten.detach().mean(dim=1) # shape (B, Heads, seq, seq) => (B, seq, seq)

        # torch.nn.Module > self.training
        # model.eval() model.train()
        #out_1 = x + torch.dropout(out, p=DROPOUT, train=self.training)
        out_1 = x + out
        out_1_norm = self.norm_1.forward(out_1)

        out_2 = self.ff.forward(out_1_norm)
        out_3 = out_1_norm + out_2
        y_prim = self.norm_2.forward(out_3)

        return y_prim, atten


class ViT(torch.nn.Module):
    def __init__(self, chw=(3, 28, 28), n_patches=14, n_blocks=3, hidden_d=64, n_heads=2):
        super().__init__()

        self.chw = chw  # (C, H, W)
        self.n_patches = n_patches
        self.patch_size = (int(chw[1] / n_patches), int(chw[2] / n_patches))
        self.hidden_d = hidden_d

        self.input_d = int(chw[0] * self.patch_size[0] * self.patch_size[1])
        self.proj = torch.nn.Linear(self.input_d, self.hidden_d)

        # 3 * 28 * 28 = 2352 pixels
        # 7*7 = 49 patches
        # 2352 / 49 = 48 pixels per patch / hidden size

        self.class_token = torch.nn.Parameter(torch.rand(1, self.hidden_d))

        self.pos_embed = torch.nn.Parameter(get_positional_embeddings(self.n_patches ** 2 + 1, self.hidden_d))
        self.pos_embed.requires_grad = False

        self.pos_enc = PositionalEncoding(self.n_patches ** 2 + 1, self.hidden_d)

        self.blocks = torch.nn.ModuleList([TransformerLayer(hidden_d, n_heads) for _ in range(n_blocks)])

        self.mlp = torch.nn.Sequential(
            torch.nn.Linear(self.hidden_d, len(dataset_full.labels)),
            torch.nn.Softmax(dim=-1)
        )

    def forward(self, x):
        n, c, h, w = x.shape

        x_patches = F.unfold(x, kernel_size=self.patch_size[0], stride=self.patch_size[0], padding=0)
        x_patches = x_patches.permute(0, 2, 1)
        x_patches = x_patches.reshape(n, self.n_patches ** 2, self.input_d)

        #x_patches = self.patches(x)
        x_proj = self.proj.forward(x_patches)

        # todo optimize
        tokens = torch.stack([torch.vstack((self.class_token, x_proj[i])) for i in range(len(x_proj))])

        # Adding positional embedding
        pos_idxs = torch.arange(0, self.n_patches ** 2 + 1).to(DEVICE)  # tensor(3) => float(3)
        out_pos = self.pos_enc.forward(pos_idxs)

        #pos_embed = self.pos_embed.repeat(n, 1, 1)
        out = tokens + out_pos.repeat(n, 1, 1)

        attens = []
        for block in self.blocks:
            out, atten = block(out)
            # max over last dim
            atten = atten[:, 1:].max(dim=-1)[0]
            #atten = atten[:, 1:].mean(dim=-1)
            attens.append(atten)

        # max atten
        atten = torch.stack(attens).max(dim=0)[0]
        atten = atten.reshape(n, self.n_patches, self.n_patches)
        # interpolate to original image size
        atten_s = F.interpolate(atten.unsqueeze(dim=1), size=(h, w), mode='bilinear', align_corners=False).squeeze(dim=1)

        # Getting the classification token only
        out = out[:, 0]

        out = self.mlp.forward(out)

        return out, atten_s



def print_model_size(model):
    total_param_size = 0
    for name, param in model.named_parameters():
        each_param_size = np.prod(param.size())
        total_param_size += each_param_size
    print(f'model size is {total_param_size}')

model = ViT()
print_model_size(model)

model = model.to(DEVICE)
optimizer = torch.optim.Adam(model.parameters(), lr=LEARNING_RATE)

metrics = {}
for stage in ['train', 'test']:
    for metric in [
        'loss',
        'acc'
    ]:
        metrics[f'{stage}_{metric}'] = []

for epoch in range(1, 100):
    for data_loader in [dataloader_train, dataloader_test]:
        metrics_epoch = {key: [] for key in metrics.keys()}

        stage = 'train'
        if data_loader == dataloader_test:
            stage = 'test'

        for x, y in tqdm(data_loader):

            x = x.to(DEVICE)
            y = y.to(DEVICE)

            y_prim, atten = model.forward(x)
            loss = torch.mean(-y * torch.log(y_prim + 1e-8))

            if data_loader == dataloader_train:
                loss.backward()
                optimizer.step()
                optimizer.zero_grad()

            np_x = x.detach().permute(0, 2, 3, 1).cpu().numpy()
            np_atten = atten.detach().cpu().numpy()
            np_y_prim = y_prim.cpu().data.numpy()
            np_y = y.cpu().data.numpy()

            idx_y = np.argmax(np_y, axis=1)
            idx_y_prim = np.argmax(np_y_prim, axis=1)

            acc = np.average((idx_y == idx_y_prim) * 1.0)

            metrics_epoch[f'{stage}_acc'].append(acc)
            metrics_epoch[f'{stage}_loss'].append(loss.cpu().item())

        metrics_strs = []
        for key in metrics_epoch.keys():
            if stage in key:
                value = np.mean(metrics_epoch[key])
                metrics[key].append(value)
                metrics_strs.append(f'{key}: {round(value, 2)}')

        print(f'epoch: {epoch} {" ".join(metrics_strs)}')

    plt.subplot(121)  # row col idx
    plts = []
    c = 0
    for key, value in metrics.items():
        plts += plt.plot(value, f'C{c}', label=key)
        ax = plt.twinx()
        c += 1

    plt.legend(plts, [it.get_label() for it in plts])

    for i, j in enumerate([4, 5, 6, 10, 11, 12, 16, 17, 18]):
        plt.subplot(3, 6, j)
        color = 'green' if idx_y[i] == idx_y_prim[i] else 'red'
        plt.title(f"pred: {dataset_full.labels[idx_y_prim[i]]}\n real: {dataset_full.labels[idx_y[i]]}", color=color)

        np_atten[i] = (np_atten[i] - np_atten[i].min()) / (np_atten[i].max() - np_atten[i].min())
        np_atten_i = np_atten[i, :, :, np.newaxis]
        plt.imshow(np_x[i] * np_atten_i)

    plt.tight_layout(pad=0.5)
    plt.show()
